'<ADbasic Header, Headerversion 001.001>
' Foldings                       = 42,46,92,155,1019,1143,1251,1546
'<Header End>
'ADBASIC-SYSTEM-FILE--V1.00--0000--0000--INC--PRO
'
'###############################################################
'#                        ADwpDIO.inc                          #
'#              INCLUDE-file for ADwin-Pro-system              #
'###############################################################
'#                                                             #
'# Contents: Definitions of functions and procedures           #
'#           to access digital modules.                        #
'#                                                             #
'#  25.10.1996  OL   creation                                  #
'#  08.04.1998  MAH  Adaption for T9 (ADbasic 3)               #
'#  23.09.1999  JK   DIG_WRITELATCH32  (problems with SYNCALL) #
'#  16.11.2001  JK   1st line : ADBASIC-SYSTEM_FILE...         #
'#  10.12.2001  MF   CO4                                       #
'#  17.01.2002  MF   C04_GetStatus                             #
'#  31.01.2002  MF   CO4                                       #
'#  23.07.2002  MF   CO4 new synclatch / dir softwarelatch     #
'#  29.07.2002  MF   DIO32                                     #
'#  02.09.2002  MG   T10                                       #
'#  06.11.2003  RRa  STORAGE                                   #
'#  15.01.2004  MF   COMP16                                    #
'#  28.01.2004  RRa  bugfix, RTC_SET()                         #
'#  11.03.2004  RRa  bugfix, MEDIA_WFR()                       #
'#  29.03.2004  RRa  bugfix, store/load Floats to/from MEDIA   #
'#  28.04.2004  RRa  SUB added: RTC_GET()                      #
'#  03.02.2005  MF   NOT T11                                   #
'#  14.04.2005  MF   ERROR WRONG PROCESSOR                     #
'#  30.11.2006  MF   bugfix PWM_ENABLE                         #
'#  09.04.2007  MF   designed for ADbasic 5                    #
'#  20.05.2008  MF   digout                                    #  
'#  23.03.2009  MF   '''+                                      #
'#                                                             #
'# Attention:                                                  #
'#   - Do not copy the first source line (containing           #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !              #
'#     Otherwise the Syntax color may be not correct.          #
'#   - This includefile requires ADbasic 4.0 or higher!        #
'#                                                             #
'###############################################################

#IF PROZESSOR=T11 THEN
#error("This include file can only be used with the processor T2, T4, T5, T8, T9 or T10.")
#ELSE

'<REGION> Declaration
'***************************************************************
' The hardware adresses depend on the CPU module 
'***************************************************************
#IF PROZESSOR = T9 THEN
#define DL         20400000H ''<HIDE>
#define DH         20400010H ''<HIDE>
#define DLX        20400020H ''<HIDE>
#define DHX        20400030H ''<HIDE>
#define D32        20400180H ''<HIDE>
#define D32I       20400198H ''<HIDE>
#define ADWR       20400060H ''<HIDE>
#define ADRD       20400070H ''<HIDE>
#define cpu_EVENT  204000F0H ''<HIDE>
#define watchdog   204001D0H ''<HIDE>
#define WD_TIMER   204001E0H ''<HIDE>
#define cpu_LED    204001F0H ''<HIDE>
#ELSE
#IF PROZESSOR = T10 THEN
#define DL         20800000H ''<HIDE>
#define DH         20800010H ''<HIDE>
#define DLX        20800020H ''<HIDE>
#define DHX        20800030H ''<HIDE>
#define D32        20800180H ''<HIDE>
#define D32I       20800198H ''<HIDE>
#define ADWR       20800060H ''<HIDE>
#define ADRD       20800070H ''<HIDE>
#define cpu_EVENT  208000F0H ''<HIDE>
#define watchdog   208001D0H ''<HIDE>
#define WD_TIMER   208001E0H ''<HIDE>
#define cpu_LED    208001F0H ''<HIDE>
#ELSE
#define DL               00H ''<HIDE>
#define DH               10H ''<HIDE>
#define DLX              20H ''<HIDE>
#define DHX              30H ''<HIDE>
#define ADWR             60H ''<HIDE>
#define ADRD             70H ''<HIDE>
#define cpu_EVENT       0F0H ''<HIDE>
#define watchdog        1D0H ''<HIDE>
#define WD_TIMER        1E0H ''<HIDE>
#define cpu_LED         1F0H ''<HIDE>
#ENDIF
#ENDIF
'<ENDREGION>

'<REGION> System

'***************************************************************
' Procedure inserts 5 nops after the poke order, if the processor
' is a T9. 
'***************************************************************

SUB poke_nop(adress,value)''<HIDE>
#IF PROZESSOR = T9 THEN
  poke(adress,value)
  nops(5)
#ELSE
#IF PROZESSOR = T10 THEN
  poke(adress,value)   
  nops(10)  
#ELSE
  poke(adress,value)
#ENDIF
#ENDIF
ENDSUB



'***************************************************************
' Procedure inserts 1 nop if the processor is a T9. 
'***************************************************************

SUB NOP_T9()''<HIDE>
#IF PROZESSOR = T9 THEN
  NOP
#ENDIF
#IF PROZESSOR = T10 THEN
  NOP
  NOP
#ENDIF
ENDSUB

'***************************************************************
' If the processor is a T9 this prcedure disables the interrupt,
' so that a process with low priority can not be interrupted.
' The procedure is used to protect a read or a write cycle.
'***************************************************************
SUB SUB_BEGIN_CRITICAL_SECTION()''<HIDE>
#IF PROZESSOR = ADSP THEN
#IF PRIORITAET = LOW
  BEGIN_CRITICAL_SECTION
#ENDIF
#ENDIF
ENDSUB

'***************************************************************
' This procedure enables the interrupt.
' (see SUB_BEGIN_CRITICAL_SECTION)
'***************************************************************
SUB SUB_END_CRITICAL_SECTION()''<HIDE>
#IF PROZESSOR = ADSP THEN
#IF PRIORITAET = LOW
  END_CRITICAL_SECTION
#ENDIF
#ENDIF
ENDSUB
'<ENDREGION>

'<REGION> DIO
'***************************************************************
''' Initialises a digital I/O module (channel 0 - 15). 
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''+     0 = input
'''+     1 = output
'''*************************************************************
SUB DigProg1(module,pattern)
  DIM adr, old as Long
  PRECALC
  adr = 32 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  old = peek( DHX )
  SUB_END_CRITICAL_SECTION
  nop_t9
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR, adr )
  nop_t9
  poke( DH , old AND 0FFFFH )
  nop_t9
  poke_nop( DLX, pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Initialises a digital I/O module (channel 16 - 31.) 
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''+     0 = input
'''+     1 = output
'''*************************************************************
SUB DigProg2(module,pattern)
  DIM adr,old as Long
  PRECALC
  adr = 32 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  old = peek( DLX )
  SUB_END_CRITICAL_SECTION
  nop_t9
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr )
  nop_t9
  poke( DL , old AND 0FFFFH )
  nop_t9
  poke_nop( DHX , pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Reads the digital inputs (channel 0 - 15). 
''' module: module address (0..255)
'''*************************************************************
FUNCTION Digin_Word1(module) AS Long
  DIM adr as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Digin_Word1 = peek(DLX)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION


'***************************************************************
''' Reads the digital inputs (channel 16 - 31). 
''' module: module address (0..255)
'''************************************************************
FUNCTION Digin_Word2(module) AS Long
  DIM adr as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Digin_Word2 = peek(DHX)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION


'***************************************************************
''' Reads the digital inputs (channel 0 - 31). 
''' module: module address (0..255)
'''*************************************************************
FUNCTION Digin_Long(module) AS Long
  DIM adr as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Digin_Long = Shift_Left(PEEK(DH),16)+PEEK(DLX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
''' Reads the digital inputs (channel 0 - 31). 
''' module: module address (0..255)
''' only for modules newer 1/2001 and ADSP CPU
'''*************************************************************
FUNCTION Digin_Long_F(module) AS Long
  DIM adr as Long
  adr = 16 OR (SHIFT_LEFT(module,8))
 
  SUB_BEGIN_CRITICAL_SECTION  
  poke_nop(ADRD,ADR)
  Digin_Long_F=PEEK(D32)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION


'***************************************************************
''' Sets the digital outputs (channel 0 - 15).
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Digout_Word1(module,pattern)
  DIM adr as Long
  PRECALC
  adr = SHIFT_LEFT(module,8)
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr )
  nop_t9
  PRECALC
  poke_nop( DLX , pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Sets the digital outputs (channel 16 - 31). 
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Digout_Word2(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 1 OR SHIFT_LEFT(module,8)
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr )
  nop_t9
  PRECALC
  poke_nop( DHX , pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Sets the digital outputs (channel 0 - 31). 
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Digout_Long(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 2 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr )
  nop_t9
  PRECALC
  poke( DH , Shift_Right(pattern,16) AND 0FFFFH )
  nop_t9
  PRECALC
  poke_nop( DLX , pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
'''Sets the digital outputs (channel 0 - 31). 
'''+ only for modules newer 1/2001 and ADSP CPU
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Digout_Long_F(module,pattern)
  DIM adr as Long
  ADR=2h OR (SHIFT_LEFT(module,8))

  SUB_BEGIN_CRITICAL_SECTION  
  POKE(ADWR,ADR)
  NOP_T9
  Poke_NOP(D32,pattern)
  SUB_END_CRITICAL_SECTION
ENDSUB

'***************************************************************
''' Reads back of the output value of the digital
'''+ outputs (channel 0 - 15). 
''' module: module address (0..255).
'''*************************************************************
FUNCTION Get_Digout_Word1(module) AS Long
  DIM adr as Long
  PRECALC
  adr = SHIFT_LEFT(module,8)
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Get_Digout_WORD1 = peek(DLX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
''' Sets and resets selected digital outputs.
''' module: module address (0..255)
''' set: Bitstring indicating the outputs which should be set
''' reset: Bitstring indicating the outputs which should be reset
'''*************************************************************
Sub Digout_Bits(module,set,reset)
  dim old as Long

  ' getting actual value and manipulating Bits
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, SHIFT_LEFT(module,8))
  old = (((Shift_Left(PEEK(DH),16)+PEEK(DLX)) or set) and not(reset))
  SUB_END_CRITICAL_SECTION

  ' writing to output
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR, SHIFT_LEFT(module,8) or 2) 
  nop_t9
  poke(DH , Shift_Right(old,16) AND 0FFFFH )
  nop_t9
  poke_nop(DLX , old AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
endsub


'***************************************************************
''' Sets and resets selected digital outputs.
'''+ only for DIO32RB and ADSP CPU
''' module: module address (0..255) 
''' set: Bitstring indicating the outputs which should be set
''' reset: Bitstring indicating the outputs which should be reset
'''*************************************************************
Sub Digout_Bits_F(module,set,reset)
  dim adr as Long
  ADR=5 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION  
  POKE(ADWR,ADR)
  NOP_T9
  Poke_NOP(D32,reset)
  SUB_END_CRITICAL_SECTION

  dec ADR

  SUB_BEGIN_CRITICAL_SECTION  
  POKE(ADWR,ADR)
  NOP_T9
  Poke_NOP(D32,set)
  SUB_END_CRITICAL_SECTION
endsub


'***************************************************************
''' Reads back of the output value of the digital
''' outputs (channel 16 - 31).
''' module: module address (0..255)
'''*************************************************************
FUNCTION Get_Digout_Word2(module) AS Long
  DIM adr as Long
  PRECALC
  adr = SHIFT_LEFT(module,8)
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Get_Digout_Word2 = peek(DHX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
''' Reads back of the output value of the digital
'''+ outputs (channel 0 - 31).
''' module: module address (0..255)
'''*************************************************************
FUNCTION Get_Digout_Long(module) AS Long
  DIM adr as Long
  PRECALC
  adr = SHIFT_LEFT(module,8)
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Get_Digout_Long = Shift_Left(PEEK(DH),16)+PEEK(DLX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
' Sets or resets one channel of a digital output.  
' module: module address (0..255)
' channel: channel (0..31)
' value: 0 = reset channel
'        1 = set channel
'***************************************************************

'SUB Digout(module,channel,value)
'  DIM adr, old, new as Long
'  PRECALC
'  adr = SHIFT_LEFT(module,8)
'  SUB_BEGIN_CRITICAL_SECTION
'  poke_nop(ADRD, adr )
'  IF (channel<16) THEN
'    old = peek(DLX)
'    SUB_END_CRITICAL_SECTION
'    IF (value=0) THEN
'      new = old AND (NOT(SHIFT_LEFT(1,channel)))
'    ELSE
'      new = old OR SHIFT_LEFT(1,channel)
'    ENDIF
'    nop_t9
'    SUB_BEGIN_CRITICAL_SECTION
'    poke(ADWR, adr)
'    nop_t9
'    poke_nop(DLX, new)
'    SUB_END_CRITICAL_SECTION
'  ELSE       
'    old = peek(DHX)
'    SUB_END_CRITICAL_SECTION
'    IF (value=0) THEN
'      new = old AND (NOT(SHIFT_LEFT(1,(channel-16))))
'    ELSE
'      new = old OR SHIFT_LEFT(1,(channel-16))
'    ENDIF
'    nop_t9
'    SUB_BEGIN_CRITICAL_SECTION
'    poke(ADWR, adr+1)
'    nop_t9
'    poke_nop(DHX, new)
'    SUB_END_CRITICAL_SECTION
'  ENDIF 
'ENDSUB


'***************************************************************
''' Sets or resets one channel of a digital output  
''' module: module address (0..255)
''' channel: channel (0..31)
''' value: 
'''+        0 = reset channel
'''+        1 = set channel
'''*************************************************************
SUB Digout(module,channel,value)
  DIM new as Long
  PRECALC
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD,SHIFT_LEFT(module,8))
  new = ((Shift_Left(PEEK(DH),16)+PEEK(DLX)) AND (NOT(SHIFT_LEFT(1,channel)))) or shift_left(value,channel)
  SUB_END_CRITICAL_SECTION
  
  if (channel<16) then
    SUB_BEGIN_CRITICAL_SECTION
    poke(ADWR,SHIFT_LEFT(module,8))
    nop_t9
    poke_nop(DLX, new AND 0FFFFH)
    SUB_END_CRITICAL_SECTION
  else
    SUB_BEGIN_CRITICAL_SECTION
    poke(ADWR,SHIFT_LEFT(module,8)+1)
    nop_t9
    poke_nop(DHX, shift_right(new,16))
    SUB_END_CRITICAL_SECTION
  endif
ENDSUB


'***************************************************************
''' Sets or resets one channel of a digital output. 
'''+only for DIO32RB and ADSP CPU
''' module: module address (0..255)
''' channel: channel (0..31)
''' value: 
'''+       0 = reset channel
'''+       1 = set channel
'''*************************************************************
SUB Digout_F(module,channel,value)
  DIM adr as Long
  dim value2 as Long

  ADR=(5h-value) OR (SHIFT_LEFT(module,8))
  value2=shift_left(1,channel)
  
  SUB_BEGIN_CRITICAL_SECTION  
  POKE(ADWR,ADR)
  NOP_T9
  Poke_NOP(D32,value2)
  SUB_END_CRITICAL_SECTION

ENDSUB


'***************************************************************
''' Latches the digital I/O module. 
''' Inputs are latched in the input registers
''' Outputs are set according to the output register.             
''' module: module address (0..255)
'''*************************************************************
SUB Dig_Latch(module)
  DIM adr as Long
  PRECALC
  adr = 22 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR, adr)
  nop_t9
  poke_nop(DLX , 0)
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Reads the input latch register (channel 0 - 15).
''' module: module address (0..255)
'''*************************************************************
FUNCTION Dig_ReadLatch1(module) AS Long
  DIM adr as Long
  PRECALC
  adr = 20 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Dig_Readlatch1 = peek(DLX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
''' Reads the input latch register (channel 16 - 31).
''' module: module address (0..255)
'''*************************************************************
FUNCTION Dig_ReadLatch2(module) AS Long
  DIM adr as Long
  PRECALC
  adr = 20 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  Dig_Readlatch2 = peek(DHX)
  SUB_END_CRITICAL_SECTION
ENDFUNCTION

'***************************************************************
''' Sets output register (channel 0 - 31).
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Dig_WriteLatch32(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 18 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr)
  nop_t9
  PRECALC
  poke( DH , Shift_Right(pattern,16) AND 0FFFFH )
  nop_t9
  PRECALC
  poke_nop( DLX , pattern AND 0FFFFH )
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Sets output register (channel 0 - 15).
'''+ATTENTION: Do not use in combination with DIG_WRITELATCH2
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''*************************************************************
SUB Dig_WriteLatch1(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 18 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr)
  nop_t9
  poke_nop( DLX , pattern)
  SUB_END_CRITICAL_SECTION
ENDSUB

'***************************************************************
''' Sets output register (channel 16 - 31)
'''+ATTENTION: Do not use in combination with DIG_WRITELATCH1
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''***************************************************************
SUB Dig_WriteLatch2(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 18 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr)
  nop_t9
  poke_nop( DHX , pattern)
  SUB_END_CRITICAL_SECTION
ENDSUB
'<ENDREGION>

'<REGION> Counter
'***************************************************************
'
' The following functions and procedures
' are for accessing counter modules. 
'
'***************************************************************

'***************************************************************
''' Activates or deactivates counter. 
''' module: module address (0..255)
''' pattern: each bit corresponds to a counter
'''*************************************************************
SUB CNT_Enable(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 20 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr )
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Resets counter. 
''' module: module address (0..255)
''' pattern: each bit corresponds to a counter
'''*************************************************************
SUB CNT_Clear(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 18 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop(DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB

'***************************************************************
''' Latches counter value to register. 
''' module: module address (0..255)
''' pattern: each bit corresponds to a counter
'''*************************************************************
SUB CNT_Latch(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Sets mode for counter.
''' module: module address (0..255)
''' pattern: each bit corresponds to a counter
'''*************************************************************
SUB CNT_SetMode(module,pattern)
  DIM adr as Long
  PRECALC
  adr = 22 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB

'***************************************************************
''' Sets mode for a counter.
''' module: module address (0..255)
''' cnt_no: number of the counter
''' mode:
'''+       0 = quadrature encoder input
'''+       1 = forward / backward counter
'''+       2 = PW input
'''*************************************************************
SUB CO4_SetMode(module,cnt_no,mode)
  DIM adr1 as Long
  DIM adr2 as Long
  DIM mode1 as Long
  DIM mode2 as Long
  DIM readback1 as Long
  DIM readback2 as Long
  PRECALC
  adr1 = 22 OR (SHIFT_LEFT(module,8))
  adr2 = adr1 + 1
  mode1 = shift_left(mode and 1, cnt_no-1)
  mode2 = shift_left(shift_right((mode and 2),1), cnt_no-1)

  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD , adr1)
  readback1 = PEEK(DLX) and not (shift_left(1, cnt_no-1)) 
  SUB_END_CRITICAL_SECTION
  nop_t9

  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD , adr2)
  readback2 = PEEK(DLX) and not( shift_left(1, cnt_no-1)) 
  SUB_END_CRITICAL_SECTION

  mode1 = mode1 or readback1  
  mode2 = mode2 or readback2  

  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr1)
  nop_t9
  poke_nop( DLX , mode1)           
  SUB_END_CRITICAL_SECTION
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr2)
  nop_t9
  poke_nop( DLX , mode2)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
'''Reads latch register 32 Bit.
'''+ remark: It is checked if a new counter value had been 
'''+         latched between reading the high-word and the
'''+         low-word. In this case it will be read again. 
'''+         (hardware latch)
''' module: module address (0..255)
''' cnt_no: number of the counter
'''*************************************************************
FUNCTION CNT_ReadLatch32(module,cnt_no) AS Long
  DIM hadr, ladr AS Long
  DIM old_hword, old_lword AS Long
  DIM hword, lword AS Long

  PRECALC
  hadr = (2*cnt_no-1) OR (SHIFT_LEFT(module,8))
  PRECALC
  ladr = (2*cnt_no-2) OR (SHIFT_LEFT(module,8))
  DO
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop(ADRD , ladr)
    old_lword = PEEK(DLX)
    SUB_END_CRITICAL_SECTION
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop(ADRD , hadr)
    old_hword = PEEK(DLX)
    SUB_END_CRITICAL_SECTION
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop(ADRD , ladr)
    lword = PEEK(DLX)
    SUB_END_CRITICAL_SECTION
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop(ADRD , hadr)
    hword = PEEK(DLX)
    SUB_END_CRITICAL_SECTION
  UNTIL ((old_lword = lword) AND (old_hword = hword))
  CNT_READLATCH32 = SHIFT_LEFT(hword, 16) + (lword AND 0FFFFH)
ENDFUNCTION

'**************************************************************
''' Reads latch register 32 Bit.
''' module: module address (0..255)
''' cnt_no: number of the counter
'''************************************************************
FUNCTION CO4_ReadLatch(module,cnt_no) AS Long
  DIM adr as Long
  adr = (cnt_no-1) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD , adr)
  CO4_READLATCH = PEEK(D32)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION


'***************************************************************
'''Determines the mode of the latch-inputs for all counters on the specified module.
'''+only for modules newer than 08/2002
'''+set mode of the 3. PWM-Register and
'''+set destination of the software latch
''' module: module address (0..255)
''' pattern: 
'''+        bit 0-3 each bit corresponds to a counter (3. PWM-Register)
'''+          0->latch with high-edge (default)
'''+          1->latch with low-edge
'''+        bit 4-7 each bit corresponds to a counter (softwarelatch)
'''+          0->latch into low-register (default)
'''+          1->latch into high-register
'''*************************************************************
SUB CO4_Set_LatchMode(module,pattern)
  DIM adr as Long
  adr = 25 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Reads latch register 16 Bit.
''' module: module address (0..255)
''' cnt_no: number of the counter
'''*************************************************************
FUNCTION CNT_ReadLatch16(module,cnt_no) AS Long
  DIM adr as Long
  IF (cnt_no<9) THEN 
    PRECALC
    adr = (2*cnt_no-2) OR (SHIFT_LEFT(module,8))
  ELSE
    PRECALC
    adr = (2*cnt_no-17) OR (SHIFT_LEFT(module,8))
  ENDIF
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  CNT_ReadLatch16 = PEEK(DLX) AND 0FFFFH           
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'***************************************************************
''' Reads counter value 32 Bit.
''' module: module address (0...255)
''' cnt_no: number of the counter
'''*************************************************************
FUNCTION CNT_Read32(module,cnt_no) AS Long
  DIM adr, hword as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , SHIFT_LEFT(1,(cnt_no-1)))   'latch counter
  SUB_END_CRITICAL_SECTION
  PRECALC
  adr = (2*cnt_no-1) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  hword = Shift_Left(PEEK(DLX),16)            'read HighWord
  SUB_END_CRITICAL_SECTION
  PRECALC
  adr = (2*cnt_no-2) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  CNT_READ32 = hword + (PEEK(DLX) AND 0FFFFH) 'read LowWord
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'**************************************************************
''' Reads counter value 32 Bit (new counter board).
''' module: module address (0..255)
''' cnt_no: number of the counter
'''************************************************************
FUNCTION CO4_Read(module,cnt_no) AS Long
  DIM adr as Long
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , SHIFT_LEFT(1,(cnt_no-1)))    'latch counter
  SUB_END_CRITICAL_SECTION
  PRECALC
  adr = (cnt_no-1) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  CO4_READ = PEEK(D32)                     'read 32 Bit-value
  SUB_END_CRITICAL_SECTION
  nop_T9
ENDFUNCTION


'***************************************************************
''' Reads counter value 16 Bit.
''' module: module address (0..255)
''' cnt_no: number of the counter
'''*************************************************************
FUNCTION CNT_Read16(module,cnt_no) AS Long
  DIM adr, hword as Long
  PRECALC
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr )
  nop_t9
  poke_nop( DLX , SHIFT_LEFT(1,(cnt_no-1)))   'latch counter
  SUB_END_CRITICAL_SECTION
  IF (cnt_no<9) THEN 
    PRECALC
    adr = (2*cnt_no-2) OR (SHIFT_LEFT(module,8))
  ELSE
    PRECALC
    adr = (2*cnt_no-17) OR (SHIFT_LEFT(module,8))
  ENDIF
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  CNT_READ16 = PEEK(DLX) AND 0FFFFH            'read LowWord
  SUB_END_CRITICAL_SECTION
ENDFUNCTION


'**************************************************************
''' Reads the status of a counter input.
''' module: module address (0..255)
''' cnt_no: number of the counter
'''************************************************************
FUNCTION CO4_GetStatus(module,cnt_no) AS Long
  DIM adr as Long
  adr = (27 + cnt_no) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr)
  CO4_GetStatus = PEEK(D32)           'read 32 Bit-value
  SUB_END_CRITICAL_SECTION
  nop_T9
ENDFUNCTION

'**************************************************************
''' Resets the error bits of the status register.
''' module: module address (0..255)
''' pattern: bitcombination; each bit corresponds with counter
'''************************************************************
SUB CO4_ResetStatus(module,pattern)
  DIM adr as Long
  adr = 27 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr)
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'**************************************************************
''' Enables the hardware clear input of a counter.
''' module: module address (0..255)
''' pattern: bitcombination; each bit corresponds with counter
'''************************************************************
SUB CO4_ClearEnable(module,pattern)
  DIM adr as Long
  adr = 24 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr)
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB

'**************************************************************
''' Enables the hardware latch input of a counter.
''' module: module address (0..255)
''' pattern: bitcombination; each bit corresponds with counter
'''************************************************************
SUB CO4_LatchEnable(module,pattern)
  DIM adr as Long
  adr = 26 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr)
  nop_t9
  poke_nop( DLX , pattern)           
  SUB_END_CRITICAL_SECTION
ENDSUB

'<ENDREGION>

'<REGION> SSI

'**************************************************************
''' Reads SSI encoder value 32 Bit.
''' module: module address (0..255)
''' channel: number of the counter
'''************************************************************
FUNCTION SSI_Read(module,channel) AS Long
  DIM adr as Long
  adr = (34 + channel) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  SSI_READ = PEEK(D32)            'read 32 Bit-value
  SUB_END_CRITICAL_SECTION
  nop_T9
ENDFUNCTION

'**************************************************************
''' Sets the number of databits for the SSI Interface.
''' module: module address (0..255)
''' channel: number of the counter
''' value: number of bits
'''************************************************************
SUB SSI_Set_Bits(module,channel,value)
  DIM adr as Long
  PRECALC
  adr = (36 + channel) OR (SHIFT_LEFT(module,8))
 
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR , adr)
  nop_t9
  poke_nop( DLX , value)           
  SUB_END_CRITICAL_SECTION
ENDSUB

'**************************************************************
''' Sets frequency of the SSI clock.
''' module: module address (0..255)
''' prescale: frequency = 10MHZ / value
'''************************************************************
SUB SSI_Set_Clock(module,prescale)
  DIM adr as Long
  PRECALC
  adr = 34 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr)
  nop_t9
  poke_nop( DLX , prescale)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'**************************************************************
''' Starts the SSI interface.
''' module: module address (0..255)
''' pattern: bitcombination each bit corresponts to one counter
'''************************************************************
SUB SSI_Start(module,pattern)
  DIM adr as Long
  Dim value2 as Long

  adr = 17 OR (SHIFT_LEFT(module,8))

  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr)
  value2 = peek(DLX)
  SUB_END_CRITICAL_SECTION
  
  value2=(pattern and 03H)or value2
   
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr)
  nop_t9
  poke_nop(DLX , value2)
  SUB_END_CRITICAL_SECTION
ENDSUB


'*************************************************************
''' Starts the SSI mode.
''' module: module address (0..255)
''' pattern: 
'''+   Bitcombination each bit corresponts to one counter
'''+       0 = single start mode
'''+       1 = continuous start mode
'''***********************************************************
sub SSI_Mode(module,pattern)
  dim adr as Long
  dim value2 as Long

  value2=pattern AND 03H
  adr = 17 OR (SHIFT_LEFT(module,8))

  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr)
  nop_t9
  poke_nop(DLX , (shift_left(value2,2)))
  SUB_END_CRITICAL_SECTION

endsub

'*************************************************************
''' Reads the SSI status.
'''+ return value 
'''+               0 = SSI value compleated
'''+               1 = SSI transfer running
''' module: module address (0..255)
''' channel: number of the SSI counter
'''***********************************************************
FUNCTION SSI_Status(module,channel) AS Long
  dim adr as Long
  dim value2 as Long
  adr = 17 OR (SHIFT_LEFT(module,8))

  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr)
  value2 = peek(DLX)
  SUB_END_CRITICAL_SECTION
 
  value2=shift_right(value2,(channel-1))
  ssi_status=value2 AND 01H
ENDFUNCTION
'<ENDREGION>

'<REGION> PWM

'***************************************************************
''' Sets and updates the PWM-4 module.
''' module: module address (0..255)
''' channel: number of the output
''' prescaler: value 0,...,7 corresponds to division by 2^0,...,2^7
''' low: Number of cycles after the prescaler for LOW time
''' high: Number of cycles after the prescaler for HIGH time
'''*************************************************************
SUB PWM_Set(module,channel,prescaler,low,high)
  DIM adr_t, adr_l, adr_h AS Long
  PRECALC
  adr_t=(channel*2-1) OR (SHIFT_LEFT(module,8))
  PRECALC
  adr_l=(channel*2+6) OR (SHIFT_LEFT(module,8))
  PRECALC
  adr_h=(channel*2-2) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_t)
  nop_t9
  poke_nop(DLX,prescaler)
  SUB_END_CRITICAL_SECTION
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_l)
  nop_t9
  poke_nop(DLX,LOW)
  SUB_END_CRITICAL_SECTION
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_h)
  nop_t9
  poke_nop(DLX,HIGH)
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Enables a output of the PWM module.
''' module: module address (0..255)
''' number: each bit corresponds to a counter
'''*************************************************************
SUB PWM_Enable(module,pattern)
  DIM adr_en, adr_clr AS Long
  PRECALC
  adr_clr=18 OR (SHIFT_LEFT(module,8))
  PRECALC
  adr_en=20 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_clr)
  nop_t9
  poke_nop(DLX,pattern)
  SUB_END_CRITICAL_SECTION
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_en)
  nop_t9
  poke_nop(DLX,pattern)
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Enables or disables external latch.  
''' module: module address (0..255)
''' cnt_select: 
'''+            1 = enable
'''+            0 = disable 
'''*************************************************************
SUB ExtLch_Enable(module,cnt_select)
  DIM adr as Long
  PRECALC
  adr = 24 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR , adr )
  nop_t9
  poke_nop( DLX , cnt_select)           
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Sets or resets an output of the PWM module.
''' module: module address (0..255)
''' channel: number of the output
'''level: 
'''+       0 = output voltage = 0V
'''+       1 = output voltage = external voltage
'''*************************************************************
SUB PWM_Out(module,channel,level)
  DIM adr_clr, adr_upd AS Long
  PRECALC
  adr_clr=18 OR (SHIFT_LEFT(module,8))
  PRECALC
  adr_upd=24 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_clr)
  nop_t9
  poke_nop(DLX,SHIFT_LEFT(1,channel-1))
  SUB_END_CRITICAL_SECTION
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr_upd)
  nop_t9
  poke_nop(DLX,SHIFT_LEFT(1,channel-1))
  SUB_END_CRITICAL_SECTION
  PWM_SET(module,channel,0,1-level,level)
ENDSUB

'<ENDREGION>

'<REGION> Storage
'*********************************************************************
' The following functions and procedures
' are for accessing the STORAGE modules. 
'*********************************************************************

'---------------------------------------------------------------
''' Reads the status word from the glue logic.
'''+ status
'''+Bit 0=1 Media in Slot
'''+Bit 1=1 Error
'''+Bit 2=1 Busy
'''+Bit 3=1 Data Can be read
'''+Bit 4=1 Data written
'''+Bit 5=1 Reset
'''+Bit 6 NA   (Might be set by other function calls)
'''+Bit 7 NA   (Might be set by other function calls)
'''+Bit 8-11
'''+                0  0  0  1 = 1d : Time-Out READY-Signal
'''+                0  0  1  0 = 2d : Time-Out WAIT-Signal
'''+                0  0  1  1 = 3d : Media was removed
'''+                0  1  0  0 = 4d : Error at READY
'''+                0  1  0  1 = 5d : Time-Out READY
'''+                0  1  1  0 = 6d : Error at DRQ (Data ReQuest)
'''+                0  1  1  1 = 7d : Time-Out DRQ (Data ReQuest)
'''+                1  0  0  0 = 8d : Media not inserted
'''+Bit 12-31 NA
''' module: module address (0..255)
'''---------------------------------------------------------------
FUNCTION Media_Stat(module) AS Long
  SUB_BEGIN_CRITICAL_SECTION  
  POKE_NOP(ADRD,3 OR SHIFT_LEFT(module,8))
  MEDIA_STAT = PEEK(D32)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION

'*********************************************************************
''' Reads a 32 bit Long value from the glue logic which was previously
'''+ read from the media.
''' module: module address (0..255)
'''*******************************************************************
FUNCTION Media_RD_DAT(module) AS Long
  SUB_BEGIN_CRITICAL_SECTION  
  POKE_NOP(ADRD,15 OR SHIFT_LEFT(module,8))
  MEDIA_RD_DAT = PEEK(D32)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION

'*********************************************************************
''' Writes an instruction word to the glue logic.
'''  module: module address (0..255)
'''  code: instruction code (Long), possible codes are:
'''+          6 - write sectors
'''+          8 - read identifier of media (256 bytes Long)
'''+          9 - read sectors
'''  sector: sector number (Long)
'''  cnt:    sector count (Long: 1...12)
'''*******************************************************************
SUB Media_WR_Cmd(module,code,sector,cnt)
  SUB_BEGIN_CRITICAL_SECTION
  POKE(ADWR,code OR SHIFT_LEFT(module,8))
  NOP_T9
  POKE_NOP(D32,(sector OR SHIFT_LEFT(cnt,28)))
  SUB_END_CRITICAL_SECTION
ENDSUB

'*********************************************************************
''' Writes one 32 bit data word to the SRAM of the glue logic.
''' module: module address (0..255)
''' daten: 32 bit data word (Long or Float)
'''*******************************************************************
SUB Media_WR_Dat(module,daten)
  SUB_BEGIN_CRITICAL_SECTION
  POKE(ADWR,7 OR SHIFT_LEFT(module,8))
  NOP_T9
  POKE_NOP(D32,daten)
  SUB_END_CRITICAL_SECTION
ENDSUB

'*********************************************************************
''' Sends a RESET command to the media for initialization or recovering
'''+ from unexpected fault conditions.
''' module: module address (0..255)
'''*******************************************************************
SUB Media_RST(module)
  SUB_BEGIN_CRITICAL_SECTION
  POKE(ADWR,20 OR SHIFT_LEFT(module,8))
  NOP_T9
  POKE_NOP(D32,0)
  SUB_END_CRITICAL_SECTION
ENDSUB

'*********************************************************************
''' Waits for respond of the media for a specified amount of time.
'''+  return value : Media status as described in MEDIA_STAT() (Long) or
'''+                 80h -> Media does not respond (time-out)
''' module: module address (0..255)
''' term: Abort condition (as described in MEDIA_STAT())
''' time: Abort time in 100s steps
'''*******************************************************************
FUNCTION Media_WFR(module,term,time) AS Long
  DIM t1, t2, t3, stat, ret AS Long
  t1=READ_TIMER()
  t3=time
#IF PROZESSOR=T10 THEN
  t3=t3*2
#ENDIF
  DO
    t2=READ_TIMER()
    stat=MEDIA_STAT(module)
    IF ((stat AND 1)=0) THEN  'media not inserted
      ret=stat
      stat=0
    ELSE
      IF ((t2-t1)>t3) THEN  'time-out
        MEDIA_RST(module)
        ret=80h
        stat=0
      ELSE
        ret=stat
      ENDIF
    ENDIF
  UNTIL ((stat AND term)=0h)
  MEDIA_WFR=ret
ENDFUNCTION

'*********************************************************************
''' Writes blocks (1...12 with 128 Longs each) of data to media.
'''+  return value: Media status As described in Media_Stat() (Long) or
'''+                 40h -> number of blocks exceeding EOF,
'''+                        no data written to media
''' module: module address (0..255)
''' f_info[]: Name of array that holds the file informations
''' file: File number (Long: 1...10)
''' s_sec: Rel. sector number of start sector in file (Long: 1...m)
''' s_cnt: Number of blocks with 128 Longs each (Long: 1...12)
''' array[]: Source array to copy data from
''' idx: First array element to copy from (Long: 1...n)
'''*******************************************************************
FUNCTION Media_WR_Blk_L(module,f_info[],file,s_sec,s_cnt,array[],idx) AS Long
  DIM i, first, last, stat, type AS Long
  last=f_info[file*2+2]
  first=f_info[file*2+1]
  IF ((last-first+2)<(s_sec+s_cnt)) THEN
    stat=40h      'Block will be written beyond EOF!!!
  ELSE
    stat=MEDIA_WFR(module,26h,150000)
    MEDIA_WR_CMD(module,6,(s_sec-1+first),s_cnt)
    FOR i=1 TO s_cnt*128
      MEDIA_WR_DAT(module,array[idx-1+i])
    NEXT i
    stat=MEDIA_STAT(module)
  ENDIF
  MEDIA_WR_BLK_L=stat
ENDFUNCTION


'---------------------------------------------------------------
''' Writes blocks (1...12 with 128 Longs each) of data to media.
'''+  return value: Media status As described in Media_Stat() (Long) or
'''+                 40h -> number of blocks exceeding EOF,
'''+                        no data written to media
''' module: module address (0..255)
''' f_info[]: Name of array that holds the file informations
''' file: File number (Long: 1...10)
''' s_sec: Rel. sector number of start sector in file (Long: 1...m)
''' s_cnt: Number of blocks with 128 Longs each (Long: 1...12)
''' array[]: Source array to copy data from
''' idx: First array element to copy from (Long: 1...n)
'''---------------------------------------------------------------
FUNCTION Media_WR_Blk(module,f_info[],file,s_sec,s_cnt,array[],idx) AS Long
  MEDIA_WR_BLK = MEDIA_WR_BLK_L(module,f_info,file,s_sec,s_cnt,array,idx)
ENDFUNCTION

'---------------------------------------------------------------
''' Writes blocks (1...12 with 128 Floats each) of data to media.
'''+  return value : Media status As described in Media_Stat() (Long) or
'''+                 40h -> number of blocks exceeding EOF,
'''+                        no data written to media
''' module: module address (0..255)
''' f_info[]: Name of array that holds the file informations
''' file: File number (Long: 1...10)
''' s_sec: Rel. sector number of start sector in file (Long: 1...m)
''' s_cnt: Number of blocks with 128 Floats each (Long: 1...12)
''' array[]: Source array to copy data from
''' idx: First array element to copy from (Long: 1...n)
'''---------------------------------------------------------------
FUNCTION Media_WR_Blk_F(module,f_info[],file,s_sec,s_cnt,array[],idx) AS Long
  DIM i, first, last, stat, type AS Long
  last=f_info[file*2+2]
  first=f_info[file*2+1]
  IF ((last-first+2)<(s_sec+s_cnt)) THEN
    stat=40h      'Block will be written beyond EOF!!!
  ELSE
    stat=MEDIA_WFR(module,26h,150000)
    MEDIA_WR_CMD(module,6,(s_sec-1+first),s_cnt)
    stat=MEDIA_WFR(module,26h,150000)
    FOR i=1 TO s_cnt*128
      MEDIA_WR_DAT(module,CAST_FloatTOLong(array[idx-1+i]))
    NEXT i
    stat=MEDIA_STAT(module)
  ENDIF
  MEDIA_WR_BLK_F=stat
ENDFUNCTION

'---------------------------------------------------------------
''' Reads blocks (1...12 with 128 Longs each) of data from media.
'''+  return value : Media status As described in Media_Stat() (Long) or
'''+                 40h -> number of blocks exceeding EOF,
'''+                        no data read from media
''' module: module address (0..255)
''' f_info[]: Name of array that holds the file informations
''' file: File number (Long: 1...10)
''' s_sec: Rel. sector number of start sector in file (Long: 1...n)
''' s_cnt: Number of blocks with 128 Longs/Floats each (Long: 1...12)
''' array[]: Destination array to copy data to
''' idx: First array element to copy To (Long: 1...n)
'''-------------------------------------------------------------
FUNCTION Media_RD_Blk_L(module,f_info[],file,s_sec,s_cnt,array[],idx) AS Long
  DIM i, first, last, stat AS Long
  last=f_info[file*2+2]
  first=f_info[file*2+1]
  IF ((last-first+2)<(s_sec+s_cnt)) THEN
    stat=40h      'Block will be read beyond EOF!!!
  ELSE
    stat=MEDIA_WFR(module,26h,150000)
    MEDIA_WR_CMD(module,9,(s_sec-1+first),s_cnt)
    stat=MEDIA_WFR(module,26h,150000)
    FOR i=1 TO s_cnt*128
      array[idx-1+i]=MEDIA_RD_DAT(module)
    NEXT i
    stat=MEDIA_STAT(module)
  ENDIF
  MEDIA_RD_BLK_L=stat
ENDFUNCTION


'---------------------------------------------------------------
''' Reads blocks (1...12 with 128 Floats each) of data from media.
'''+  return value : Media status As described in Media_Stat() (Long) or
'''+                40h -> number of blocks exceeding EOF,
'''+                       no data read from media
''' module: module address (0..255)
''' f_info[]: Name of array that holds the file informations
''' file: File number (Long: 1...10)
''' s_sec: Rel. sector number of start sector in file (Long: 1...n)
''' s_cnt: Number of blocks with 128 Longs/Floats each (Long: 1...12)
''' array[]: Destination array to copy data to
''' idx: First array element to copy To (Long: 1...n)
'''--------------------------------------------------------------
FUNCTION Media_RD_Blk_F(module,f_info[],file,s_sec,s_cnt,array[],idx) AS Long
  DIM i, first, last, stat AS Long
  last=f_info[file*2+2]
  first=f_info[file*2+1]
  IF ((last-first+2)<(s_sec+s_cnt)) THEN
    stat=40h      'Block will be read beyond EOF!!!
  ELSE
    stat=MEDIA_WFR(module,26h,150000)
    MEDIA_WR_CMD(module,9,s_sec-1+first,s_cnt)
    stat=MEDIA_WFR(module,26h,150000)
    FOR i=1 TO s_cnt*128
      array[idx-1+i]=CAST_LongTOFloat(MEDIA_RD_DAT(module))
    NEXT i
    stat=MEDIA_STAT(module)
  ENDIF
  MEDIA_RD_BLK_F=stat
ENDFUNCTION


'---------------------------------------------------------------
''' Resets media and reads from sector 2 the file information table,
'''+ where all start and end sectors of the data files are stored.
'''+ Unable to read sector 2 if RESET,BUSY,ERROR,no media in slot!
'''+  return value : Media status As described in Media_Stat() (Long)
''' module: module address (0..255)
''' array[]: Name of array (>= 22 elements) to store start & end sectors
'''---------------------------------------------------------------
FUNCTION Media_RD_FileInfo(module,array[]) AS Long
  DIM n, stat AS Long
  stat=MEDIA_STAT(module)
  IF ((stat AND 1h)=1h) THEN
    MEDIA_RST(module)
    stat=MEDIA_WFR(module,26h,1200)
    MEDIA_WR_CMD(module,9,2,1)
    stat=MEDIA_WFR(module,26h,150000)
    FOR n=1 TO 22
      array[n]=MEDIA_RD_DAT(module)
    NEXT n
  ENDIF
  MEDIA_RD_FILEINFO=stat
ENDFUNCTION
'<ENDREGION>

'<REGION> Real-time clock

'********************** Real-Time-Clock (RTC): ***********************
'The 32 bit value holds the information for year, month, day,
'hour, minute and second in the following manner:
'   bits: | 31:26 | 25:22 | 21:17 | 16:12 | 11:06 | 05:00 |
'   ------+-------+-------+-------+-------+-------+-------+
'         | year  | month |  day  | hour  |minute |second |

'*********************************************************************
' Writes the coded 32 bit word containing the date and time infor-
' mation to the RTC.
' module: module address (0..255)
'*********************************************************************
SUB RTC_WR(module,value) ''<HIDE>
  SUB_BEGIN_CRITICAL_SECTION
  POKE(ADWR,16 OR SHIFT_LEFT(module,8))
  NOP_T9
  POKE_NOP(D32,value)
  SUB_END_CRITICAL_SECTION
ENDSUB

'*********************************************************************
' Reads the coded 32 bit word containing the date and time information
' from the RTC.
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_RD(module) AS Long ''<HIDE>
  SUB_BEGIN_CRITICAL_SECTION
  POKE_NOP(ADRD,17 OR SHIFT_LEFT(module,8))
  RTC_RD=PEEK(D32)
  SUB_END_CRITICAL_SECTION
  NOP_T9
ENDFUNCTION


'---------------------------------------------------------------
''' Sets the RTC to the desired value. if any of the entries are not
'''+ valid neither value will be accepted and the RTC remains unchanged!
''' module: module address (0..255)
''' year:   Last two digits of the year 0...63 (Long)
''' month:  1...12 (Long)
''' day:    1...31 (Long)
''' hour:   0...23 (Long)
''' minute: 0...59 (Long)
''' second: 0...59 (Long)
'''-------------------------------------------------------------
SUB RTC_Set(module,year,month,day,hour,min,sec)
  DIM val AS Long
  val=0
  val=SHIFT_LEFT(year AND 3Fh,26)            'Bits 31:26
  val=val OR SHIFT_LEFT(month AND 0Fh,22)    'Bits 25:22
  val=val OR SHIFT_LEFT(day AND 1Fh,17)      'Bits 21:17
  val=val OR SHIFT_LEFT(hour AND 1Fh,12)    'Bits 16:12
  val=val OR SHIFT_LEFT(min AND 3Fh,6)      'Bits 11:06
  val=val OR (sec AND 3Fh)                  'Bits 05:00
  RTC_WR(module,val)
ENDSUB


'---------------------------------------------------------------
''' Reads out the current date and time from the RTC at once.
'''+ Parameters are preferably Longs.
''' module: module address (0..255)
''' year:   Last two digits of the year 0...63 (Long)
''' month:  1...12 (Long)
''' day:    1...31 (Long)
''' hour:   0...23 (Long)
''' minute: 0...59 (Long)
''' second: 0...59 (Long)
'''-------------------------------------------------------------
SUB RTC_Get(module,year,month,day,hour,min,sec)
  DIM val AS Long
  val=RTC_RD(module)
  year=SHIFT_RIGHT(val,26) AND 3Fh    'Bits 31:26
  month=SHIFT_RIGHT(val,22) AND 0Fh    'Bits 25:22
  day=SHIFT_RIGHT(val,17) AND 1Fh      'Bits 21:17
  hour=SHIFT_RIGHT(val,12) AND 1Fh    'Bits 16:12
  min=SHIFT_RIGHT(val,6) AND 3Fh      'Bits 11:06
  sec=val AND 3Fh                     'Bits 05:00
ENDSUB

'*********************************************************************
' Reads out the current year from the RTC (2000...2063).
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_YEAR(module) AS Long ''<HIDE>
  RTC_GET_YEAR=SHIFT_RIGHT(RTC_RD(module),26)+2000
ENDFUNCTION

'*********************************************************************
' Reads out the current month from the RTC (1...12).
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_MONTH(module) AS Long ''<HIDE>
  RTC_GET_MONTH=SHIFT_RIGHT(RTC_RD(module),22) AND 0Fh
ENDFUNCTION

'*********************************************************************
' Reads out the current day from the RTC.
' (value ranges from 1...28/29/30/31 depending on month and leapyear)
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_DAY(module) AS Long ''<HIDE>
  RTC_GET_DAY=SHIFT_RIGHT(RTC_RD(module),17) AND 1Fh
ENDFUNCTION

'*********************************************************************
' Reads out the current hour from the RTC (0...23, military notation).
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_HOUR(module) AS Long ''<HIDE>
  RTC_GET_HOUR=SHIFT_RIGHT(RTC_RD(module),12) AND 1Fh
ENDFUNCTION

'*********************************************************************
' Reads out the current minute from the RTC (0...59).
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_MINUTE(module) AS Long ''<HIDE>
  RTC_GET_MINUTE=SHIFT_RIGHT(RTC_RD(module),6) AND 3Fh
ENDFUNCTION

'*********************************************************************
' Reads out the current second from the RTC (0...59).
' module: module address (0..255)
'*********************************************************************
FUNCTION RTC_GET_SECOND(module) AS Long ''<HIDE>
  RTC_GET_SECOND=RTC_RD(module) AND 3Fh
ENDFUNCTION
'<ENDREGION>

'<REGION> Comp-16

'*************************************************************
' COMP 16
'*************************************************************
''' Determines the lower and upper threshold value for a specified channel.
''' module: module address (0..255)
''' channel: channel (1-16)
''' value_high: switch point to set output bit to 1
''' value_low: switch point to set output bit to 0
'''***********************************************************
SUB Comp_Set(module,channel,value_high,value_low)
  DIM adr AS Long
  DIM CH,HIGH,LOW as Long
 
  adr=20 OR (SHIFT_LEFT(module,8))
  CH=shift_left(((channel-1) AND 0FH),20) 
  HIGH=shift_left((value_high AND 3FFH),10)
  LOW=value_low AND 3FFH
 
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr)
  nop_t9
  poke_nop(D32,(CH OR HIGH OR LOW))
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Reads the digital inputs (channel 0 - 15). 
''' module: module address (0..255)
'''*************************************************************
FUNCTION Comp_Digin_Word(module) AS Long
  DIM adr as Long
  adr = 16 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr )
  COMP_DIGIN_WORD = peek(DLX)
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION

'***************************************************************
''' Reads the digital inputs (channel 0 - 7) by diff.  
''' module: module address (0..255)
'''*************************************************************
FUNCTION Comp_Digin_Word_Diff(module) AS Long
  dim adr as Long
  adr = 23 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr)
  COMP_DIGIN_WORD_DIFF = peek(D32)
  SUB_END_CRITICAL_SECTION
  nop_T9
ENDFUNCTION

'*************************************************************
''' Reads the analog value, min value or max value
''' module: module address (0..255).
''' channel: number of channel (1-16)
''' select_val: 
'''+             0 = read value
'''+             1 = read min value
'''+             2 = read max value
'''***********************************************************
FUNCTION Comp_Read(module,channel,select_val) AS Long
  dim adr as Long
  dim value2 as Long
  adr = (23+channel) OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRD, adr)
  value2 = peek(D32)
  SUB_END_CRITICAL_SECTION
 
  COMP_READ=(shift_right(value2,(10*select_val))) AND 3FFH
ENDFUNCTION

'*************************************************************
''' Resets min / max value (max=0 min=1024)
''' module: module address (0..255)
''' pattern: each bit corresponds to a channel
'''***********************************************************
SUB Comp_Reset(module,pattern)
  DIM adr AS Long
  adr=21 OR (SHIFT_LEFT(module,8))
  
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr)
  nop_t9
  poke_nop(D32,(pattern))
  SUB_END_CRITICAL_SECTION
ENDSUB

'*************************************************************
''' Selects the channels to write to the internal FIFO
''' module: module address (0..255)
''' pattern: 0= channel 1/2  1= channel 3/4 ...
'''***********************************************************
sub Comp_FIFO_Select(module,pattern)
  DIM adr AS Long
  adr=17 OR (SHIFT_LEFT(module,8))
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr)
  nop_t9
  poke_nop(D32,(pattern+1))
  SUB_END_CRITICAL_SECTION
endsub

'---------------------------------------------------------------
''' Reads the last 2 x 1 024 measurement values of a channel pair from the internal FIFO memory and transfers the values to 2 arrays.
''' module: module address (0..255)
''' arrayLowChannel[]: values of low channels
''' arrayHighChannel[]: values of high channels
'''---------------------------------------------------------------
sub Comp_FIFO_Read(module,arrayLowChannel[],arrayHighChannel[])
  DIM adr AS Long
  dim i as Long
  dim value as Long
  dim select as Long

  adr=18 OR (SHIFT_LEFT(module,8))

  'read select for new start
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop(ADRd,adr-1)
  nop_t9
  select=peek(Dlx)
  SUB_END_CRITICAL_SECTION
  nop_t9


  'stop write to fifo
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr-1)
  nop_t9
  poke_nop(D32,(0))
  SUB_END_CRITICAL_SECTION

  'set adress
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr)
  nop_t9
  poke_nop(D32,0)
  SUB_END_CRITICAL_SECTION

  'read value
  for i=1 to 1024
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop(ADRD, adr-4)
    value = peek(D32)
    SUB_END_CRITICAL_SECTION
    arrayHighChannel[i]=value AND 0FFFFH
    arrayLowChannel[i]=shift_right(value,16)      
  next

  'start write to fifo
  SUB_BEGIN_CRITICAL_SECTION
  poke(ADWR,adr-1)
  nop_t9
  poke_nop(D32,select)
  SUB_END_CRITICAL_SECTION
endsub

'<ENDREGION>
#ENDIF

