'<ADbasic Header, Headerversion 001.001>
' Foldings                       = 40,120,183,275,331
'<Header End>
'ADBASIC-SYSTEM-FILE--V1.00--0000--0000--INC--PRO
'
'###############################################################
'#                        ADwinPRO.inc                         #
'#              INCLUDE-file for ADwin-Pro-system              #
'###############################################################
'#                                                             #
'# Contents: Definition of functions and procedures            #
'#           to access ADwin PRO SYSTEM                        #
'#                                                             #
'#  25.10.1996       creation                                  #
'#  15.01.1997       Befehle fr den Watchdog eingefhrt       #
'#  08.04.1998  OL   changes for T9/ADSP (ADbasic 3)           #
'#  16.11.2001  JK   comments in English                       #
'#  16.11.2001  JK   SUB_BEGIN/END_CRITICAL_SECTION            #
'#  16.11.2001  JK   1st line : ADBASIC-SYSTEM_FILE...         #
'#  02.09.2002  MG   T10                                       #
'#  05.10.2002  MF   T10                                       #
'#  17.10.2003  MF   DIGIN 0                                   # 
'#  29.09.2004  MF   Bugfix Checkled                           #
'#  03.02.2005  MF   NOT T11                                   #
'#  14.04.2005  MF   ERROR WRONG PROCESSOR                     #
'#  08.05.2007  MF   designed fot ADbasic 5                    #
'#  20.03.2009  MF   '''+                                      #
'#  10.08.2010  MF   syntax_highlight AD,DA,DIO,EXT            #
'#  30.04.2014  GC   inserted regions                          #
'#                                                             #
'# Attention:                                                  #
'#   - Do not copy the first source line (containing           #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !              #
'#     Otherwise the Syntax color may be not correct.          #
'#   - This includefile requires ADbasic 4.0 or higher!        #
'#                                                             #
'###############################################################

#IF PROZESSOR=T11 THEN
#error("This include file can only be used with the processor T2, T4, T5, T8, T9 or T10.")
#else

'<REGION> Declaration
'---------------------------------------------------------------
' Base-address for the different types of Pro-modules.
'---------------------------------------------------------------
'''system constant which can be used by:
'''SetLED,CheckLED
'''EventEnable,EventSource
'''SyncEnable,SyncStat
#DEFINE DIO 000h  'DIO - MASK 


'''system constant which can be used by:
'''SetLED,CheckLED
'''EventEnable,EventSource
'''SyncEnable,SyncStat
#DEFINE AD  040h  'A/D - MASK 


'''system constant which can be used by:
'''SetLED,CheckLED
'''EventEnable,EventSource
'''SyncEnable,SyncStat
#DEFINE DA  080h  'D/A - MASK 


'''system constant which can be used by:
'''SetLED,CheckLED
#DEFINE CPU 0A0h  'CPU - MASK 

'''system constant which can be used by:
'''SetLED,CheckLED
'''EventEnable,EventSource
'''SyncEnable,SyncStat
#DEFINE EXT 0C0h  'EXT - MASK 


'---------------------------------------------------------------
' Address-table to access the hardware of the Pro-system,
' depending on the CPU in use (ADSP OR Transputer).
'---------------------------------------------------------------
#IF PROZESSOR = T9 THEN
#define DL         20400000h ''<HIDE>
#define DH         20400010h ''<HIDE>
#define DLX        20400020h ''<HIDE>
#define DHX        20400030h ''<HIDE>
#define ADWR       20400060h ''<HIDE>
#define ADRD       20400070h ''<HIDE>
#define CPU_EVENT  204000F0h ''<HIDE>
#define WATCHDOG   204001D0h ''<HIDE>
#define WD_TIMER   204001E0h ''<HIDE>
#define CPU_LED    204001F0h ''<HIDE>
#define CPU_DIG_ADR 204001B0H ''<HIDE> 
#ELSE
#IF PROZESSOR = T10 THEN
#define DL         20800000h ''<HIDE>
#define DH         20800010h ''<HIDE>
#define DLX        20800020h ''<HIDE>
#define DHX        20800030h ''<HIDE>
#define ADWR       20800060h ''<HIDE>
#define ADRD       20800070h ''<HIDE>
#define CPU_EVENT  208000F0h ''<HIDE>
#define WATCHDOG   208001D0h ''<HIDE>
#define WD_TIMER   208001E0h ''<HIDE>
#define CPU_LED    208001F0h ''<HIDE>
#define CPU_DIG_ADR 208001B0H ''<HIDE> 
#ELSE
#define DL               00h ''<HIDE>
#define DH               10h ''<HIDE>
#define DLX              20h ''<HIDE>
#define DHX              30h ''<HIDE>
#define ADWR             60h ''<HIDE>
#define ADRD             70h ''<HIDE>
#define CPU_EVENT       0F0h ''<HIDE>
#define WATCHDOG        1D0h ''<HIDE>
#define WD_TIMER        1E0h ''<HIDE>
#define CPU_LED         1F0h ''<HIDE>
#ENDIF
#ENDIF
'<ENDREGION> 

'<REGION> Helper functions
'---------------------------------------------------------------
' Procedure which adds five NOP-instructions after a REG.-WRITE
' to compensate delaytimes.
'---------------------------------------------------------------
SUB poke_nop(adress,value) ''<HIDE>
#IF PROZESSOR = T9 THEN
  poke(adress,value)
  nops(5)   
#ELSE
#IF PROZESSOR = T10 THEN
  poke(adress,value)
  nops(10)
#ELSE
  poke(adress,value)
#ENDIF
#ENDIF
ENDSUB

'---------------------------------------------------------------
' Procedure which adds one additional NOP-instruction only for
' the ADSP processor, because only necessary there.
'---------------------------------------------------------------
SUB NOP_T9() ''<HIDE>
#IF PROZESSOR = T9 THEN
  NOP
#ENDIF
#IF PROZESSOR = T10 THEN
  NOP
  NOP
#ENDIF
ENDSUB


'---------------------------------------------------------------
' If a low-priority process accesses a Pro-module it must not be
' interrupted by any other process !!!
'---------------------------------------------------------------
SUB SUB_BEGIN_CRITICAL_SECTION() ''<HIDE>
#IF PROZESSOR = ADSP THEN
#IF PRIORITAET = LOW
  BEGIN_CRITICAL_SECTION
#ENDIF
#ENDIF
ENDSUB

SUB SUB_END_CRITICAL_SECTION()''<HIDE>
#IF PROZESSOR = ADSP THEN
#IF PRIORITAET = LOW
  END_CRITICAL_SECTION
#ENDIF
#ENDIF
ENDSUB

'---------------------------------------------------------------
' calculates address value
'---------------------------------------------------------------
FUNCTION FUNCTIONS_WERT(BEFEHL,module,TYP) AS Long ''<HIDE>
  FUNCTIONS_WERT = (BEFEHL OR TYP) OR (SHIFT_LEFT(module,8))
ENDFUNCTION

'<ENDREGION> 

'<REGION> System
'***************************************************************
''' Switches a LED on or off.
''' module: module address (0..255)
''' moduletype: type of module (AD,DA,DIO,EXT,CPU)
''' value: 
'''+   0 off
'''+   1 on
'''*************************************************************
SUB SetLED(module,moduletype,value)
  DIM adr as Long
  IF (moduletype=cpu) THEN
    POKE(CPU_LED, value)
    nop_t9
  ELSE
    PRECALC
    adr = FUNCTIONS_WERT( 42, module , moduletype)
    SUB_BEGIN_CRITICAL_SECTION
    poke( ADWR, adr )
    nop_t9
    poke_nop( DLX, value)
    SUB_END_CRITICAL_SECTION
  ENDIF
ENDSUB


'***************************************************************
''' Checks the LED on a module.
''' module: module address (0..255)
''' moduletype: type of module (AD,DA,DIO,EXT,CPU)
'''*************************************************************
FUNCTION CheckLED(module,moduletype) AS Long
  DIM adr as integer

  IF (moduletype=cpu) THEN
    CHECKLED = (PEEK(CPU_LED) and 1)
    nop_t9
  ELSE
    PRECALC
    adr = FUNCTIONS_WERT( 42, module , moduletype)
    SUB_BEGIN_CRITICAL_SECTION
    poke_nop( ADRD, adr )
    CHECKLED = peek( DLX) AND 1
    SUB_END_CRITICAL_SECTION
    nop_t9
  ENDIF
ENDFUNCTION


'***************************************************************
''' Activates or deactivates the external Event input.
''' module: module address (0..255)  
''' moduletype: type of module (AD,DA,DIO,EXT
''' value: 
'''+       0  deactivate 
'''+       1  activate
'''*************************************************************
SUB EventEnable(module,moduletype,value)
  DIM adr as Long
  POKE( cpu_EVENT, 256*value)    'enable the EVENT on the processormodule
  PRECALC
  adr = FUNCTIONS_WERT( 56, module , moduletype)
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr )
  nop_t9
  poke_nop( DLX, 9-value)    ' EVENT on module 
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
'''  Defines whether rising/falling edge is triggering the Event.
'''  module: module address (0..255)   
'''  moduletype: type of module (AD,DA,DIO,EXT)
'''  value:  
'''+  0  rising edge             
'''+  1  rising edge             
'''+  2  falling edge             
'''+  3  rising and falling edge  
''' ************************************************************
SUB EventSource(module,moduletype,value)
  DIM adr as Long
  PRECALC
  adr = FUNCTIONS_WERT( 59, module , moduletype)
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr )
  nop_t9
  poke_nop( DLX, value)    'EVENTSOURCE on module 
  SUB_END_CRITICAL_SECTION
ENDSUB
'<ENDREGION> 

'<REGION> Sync

'***************************************************************
''' Activates or deactivates the Sync-function of a module.
''' module: module address (0..255) 
''' moduletype: type of module (AD,DA,DIO,EXT)
''' enable: 
'''+      0 deactivate 
'''+      1 activate
'''*************************************************************
SUB SyncEnable(module,moduletype,enable)
  DIM adr AS Long
  PRECALC
  adr = FUNCTIONS_WERT( 58, module , moduletype)
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, adr )
  nop_t9
  poke_nop( DLX, enable)
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Latches all modules which have been activated with the 
'''+ SYNCENABLE command.
'''*************************************************************
SUB SyncAll()
  SUB_BEGIN_CRITICAL_SECTION
  poke( ADWR, 52)
  nop_t9
  poke_nop( DLX, 255)
  SUB_END_CRITICAL_SECTION
ENDSUB


'***************************************************************
''' Checks if the Sync-function of a module is activated.
'''+  result:   
'''+       0 deactivated   
'''+       1 activated
'''+       3 enabled for burst-measurement sequences (only for the modules Pro-AIn-F-x/14).
'''  module:   module-address (0..255)     
'''  moduletype: type of module (AD,DA,DIO,EXT)
'''*************************************************************
FUNCTION SyncStat(module,moduletype) AS Long
  DIM adr as Long
  PRECALC
  adr = FUNCTIONS_wert( 58, module , moduletype)
  SUB_BEGIN_CRITICAL_SECTION
  poke_nop( ADRD, adr )
  SyncStat  = peek( DLX) AND 1
  SUB_END_CRITICAL_SECTION
  nop_t9
ENDFUNCTION
'<ENDREGION> 

'<REGION> Watchdog
'***************************************************************
' commands for processor module 
'***************************************************************

'***************************************************************
''' Activates watchdog. 
'''*************************************************************
SUB StartWatchdog()
  poke(watchdog,1)
  nop_t9
ENDSUB

'***************************************************************
''' Deactivates watchdog.
'''*************************************************************
SUB StopWatchdog()
  poke(watchdog,0)
  nop_t9
ENDSUB

'***************************************************************
''' Resets the watchdog timer. 
'''+ If the watchdog is activated, the user program has to call  
'''+ this function at least every 80ms.
'''+ For ADSP processor modules produced in year 2000 or later 
'''+ the minimum time interval is 1600 ms.
'''*************************************************************
SUB ResetWatchdogTimer()
  poke(WD_TIMER,1)
  nop_t9
ENDSUB
'<ENDREGION> 

'***************************************************************
''' Checks whether the DIG_IN-function of a CPU module is activated.
'''+ result:  0/1  activated/deactivate
'''+ activated = a falling edge has been received 
'''*************************************************************
FUNCTION CPU_Digin() AS Long
  CPU_DIGIN = peek(cpu_DIG_ADR) AND 1
ENDFUNCTION

#ENDIF
