'<TiCoBasic Header, Headerversion 001.001>
' Foldings                       = 30,108
'<Header End>
'ADBASIC-SYSTEM-FILE--V1.00--0000--0000--INC--PRO
'<SMARTFORMAT OFF> do not remove !!!
'
'##############################################################################
'#                        DIO32Tico.inc                                       #
'#              INCLUDE-file for DIO 32 with Tico                             #
'##############################################################################
'#                                                                            #
'#                                                                            #
'#  08.08.2013  mf   creation                                                 #
'#  11.10.2016  mf   bugfix + optimize                                        #
'#  12.10.2016  mf   new sub ls_reset                                         #
'#  04.11.2016  mf   Calc_Processdelay                                        #
'#  25.01.2021  mf   Calc_TicksToNs                                           #
'#                                                                            #
'#                                                                            #
'# Attention:                                                                 #
'#   - Do not copy the first source line (containing                          #
'#     "'ADBASIC-SYSTEM-FILE...") to user files !                             #
'#     Otherwise the Syntax color may be not correct.                         #
'#   - This includefile requires ADbasic 5 or higher!                         #
'#                                                                            #
'##############################################################################

#if adwin_system=ADwin_ProII
#else
#error("This include file can only be used with the ADwin ProII")  
#endif

'<REGION> System

'***************************************************************
''' Converts a process frequency [Hz] into processor ticks (processdelay or cycle time).
''' frequency:  frequency[Hz]
'''*************************************************************
Function Calc_Processdelay(frequency) as long
  
#if prozessor=tico1
  calc_processdelay=50000000/(frequency)
#else
  calc_processdelay=100000000/(frequency)
#endif
  
endfunction



'***************************************************************
''' Converts timer ticks into ns
''' ticks: counter value ticks into ns
'''*************************************************************
Function Calc_TicksToNs(ticks) As long
  Calc_TicksToNs=(ticks)*20
endfunction

'---------------------------------------------------------------
' If a low-priority process accesses a Pro-module it must not be
' interrupted by any other process !!!
'---------------------------------------------------------------

Sub Sub_Begin_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  Begin_Critical_Section
#EndIf
EndSub

Sub Sub_End_Critical_Section() ''<HIDE>
#If PRIORITAET = LOW
  End_Critical_Section
#EndIf
EndSub

'******************************************************************************
'''Switches the LED on or off.
'''value: 
'''+  0 off
'''+  1 on
'''****************************************************************************
SUB Set_LED(value)
  out (80804000H, value)
ENDSUB

'******************************************************************************
'''Checks the LED.
'''+Return:
'''+   0 off
'''+   1 on
'''****************************************************************************
FUNCTION Check_LED() AS Long
  Check_LED=in(80804000h)
ENDFUNCTION

'*******************************************************************************
'''Creates an event for the master cpu.
'''*****************************************************************************
SUB Trigger_Event()
  out (108041E0h, 1)
ENDSUB

'*******************************************************************************
'''Creates an event for the slave module.
'''*****************************************************************************
sub Trigger_Sync()
  out (1080490Ah,1)
endsub
'<ENDREGION>

'<REGION> LS-BUS

'---------------------------------------------------------------
' Function only For internal use!
' receive LS-Bus
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
' buf[]   : array
' return  : error code (0->ok)
'---------------------------------------------------------------
Function I_Receive(channel, dev_adr, reg_adr, cnt_dat, buf[]) As Long ''<HIDE>
  Dim cnt_i, frc, ftc, tmout, tmstart, th As Long
  Dim ret As Long

  do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  until (ftc < 28)                                          ' platz zum senden
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(cnt_dat - 1, 4) Or dev_adr))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, reg_adr)              ' Header 2
  Sub_End_critical_section
  sleep(2)
  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc = 0)
  th = 520                                                  ' first 6.75 mikrosec
  For cnt_i = 1 To cnt_dat                                  ' data count
    tmstart = Read_Timer()
    Do                                                      ' rx wait
      tmout = Read_Timer() - tmstart
      frc = IN(00001012h + (channel - 1) * 16)
    Until ((frc >= 1) Or (tmout > th))
    
    
    nop
    nop
    nop
    buf[cnt_i] = IN(00001010h + (channel - 1) * 16)
    
   
    th = 350                                                '  6 mikrosec
  Next cnt_i
  tmstart = Read_Timer()
  Do                                                        ' ack wait
    tmout = Read_Timer() - tmstart                          ' rx count
    frc = in(00001012h + (channel - 1) * 16)
  Until ((frc >= 1) Or (tmout > 320))                       ' 6 mikrosec '1800
  If (frc >= 1) Then
    nop
    
    ret = IN(00001010h + (channel - 1) * 16)
  Else
    ret = -1
  EndIf
  I_Receive = ret
EndFunction

'---------------------------------------------------------------
' Function only For internal use!
' receive LS-Bus
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' value32 : return value (32 bit) 
' return  : error code (0->ok)
'---------------------------------------------------------------
Function I_Receive32(channel, dev_adr, reg_adr,value32) As Long ''<HIDE>
  Dim cnt_i, frc, ftc, tmout, tmstart, th As Long
  
  dim inhelp as long

  do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  until (ftc < 28)                                          ' platz zum senden
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(4 - 1, 4) Or dev_adr))  ' Header 1
  inhelp=0 'nop
  out(00001010h + (channel - 1) * 16, reg_adr)              ' Header 2
  Sub_End_critical_section
  sleep(2)
    
  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc = 0)
  th = 520                                                  ' first 6.75 mikrosec
  For cnt_i = 0 To 24 step 8                                     ' data count
    tmstart = Read_Timer()
    Do                                                      ' rx wait
      tmout = Read_Timer() - tmstart
      frc = IN(00001012h + (channel - 1) * 16)
    Until ((frc >= 1) Or (tmout > th))
    
    
    nop
    nop
    inhelp= inhelp or shift_left(IN(00001010h + (channel - 1) * 16),cnt_i)
    
   
    th = 350                                                '  6 mikrosec
  Next cnt_i
  tmstart = Read_Timer()
  value32=inhelp
  Do                                                        ' ack wait
    tmout = Read_Timer() - tmstart                          ' rx count
    frc = in(00001012h + (channel - 1) * 16)
  Until ((frc >= 1) Or (tmout > 320))                       ' 6 mikrosec '1800
  If (frc >= 1) Then
    I_Receive32 = IN(00001010h + (channel - 1) * 16)
  Else
    I_Receive32 = -1
  EndIf
EndFunction

'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
' buf[]   : array
' return  : error code (0->ok)
'---------------------------------------------------------------
Function I_Send(channel, dev_adr, reg_adr, cnt_dat, buf[]) As Long ''<HIDE>
  Dim cnt_i, frc, ftc, tmout, tmstart As Long
  Dim ret As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - cnt_dat))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(cnt_dat - 1, 4) Or dev_adr))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop
  For cnt_i = 1 To cnt_dat                                  ' data count
    out(00001010h + (channel - 1) * 16, buf[cnt_i])         '
  Next cnt_i
  Sub_End_critical_section
  sleep(2)
  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc = 0)
  tmstart = Read_Timer()
  Do                                                        ' ack wait
    tmout = Read_Timer() - tmstart
    frc = IN(00001012h + (channel - 1) * 16)
  Until ((frc >= 1) Or (tmout > 400))                       ' 6 mikrosec
  If (frc >= 1) Then
    ret = IN(00001010h + (channel - 1) * 16)
  Else
    ret = -1
  EndIf
  I_Send = ret
EndFunction


'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' value   : outvalue 32 bit
' return  : error code (0->ok)
'---------------------------------------------------------------
Function I_Send32(channel, dev_adr, reg_adr, value) As Long ''<HIDE>
  Dim cnt_i, frc, ftc, tmout, tmstart As Long
  Dim ret As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - 4))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(4 - 1, 4) Or dev_adr))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop
  For cnt_i = 0 To 24 step 8                                    ' data count
    out(00001010h + (channel - 1) * 16, shift_right(value,cnt_i) AND 0FFH)         '
  Next cnt_i
  Sub_End_critical_section
  sleep(2)
  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc = 0)
  tmstart = Read_Timer()
  Do                                                        ' ack wait
    tmout = Read_Timer() - tmstart
    frc = IN(00001012h + (channel - 1) * 16)
  Until ((frc >= 1) Or (tmout > 400))                       ' 6 mikrosec
  If (frc >= 1) Then
    nop
    I_Send32 = IN(00001010h + (channel - 1) * 16)
  Else
    I_Send32 = -1
  EndIf
 
EndFunction


'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus no acknowlege
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
' buf[]   : array
'---------------------------------------------------------------
Sub I_Send_NA(channel, dev_adr, reg_adr, cnt_dat, buf[]) ''<HIDE>
  Dim cnt_i, ftc As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - cnt_dat))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(cnt_dat - 1, 4) Or dev_adr))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop
  For cnt_i = 1 To cnt_dat                                  ' data count
    out(00001010h + (channel - 1) * 16, buf[cnt_i])         '
  Next cnt_i
  
  Sub_End_critical_section
EndSub

'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus no acknowlege
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' value: outvalue
'---------------------------------------------------------------
Sub I_Send_NA32(channel, dev_adr, reg_adr, value) ''<HIDE>
  Dim cnt_i, ftc As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - 4))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(4 - 1, 4) Or dev_adr))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop

  For cnt_i = 0 To 24 step 8                                    ' data count
    out(00001010h + (channel - 1) * 16, shift_right(value,cnt_i) AND 0FFH)         '
  Next cnt_i
  
  
  Sub_End_critical_section
EndSub

'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus broadcast no acknowlege
' module  : module address
' channel : number of the LS interface on the LS module
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
' buf[]   : array
'---------------------------------------------------------------
Sub I_Send_BC(channel, reg_adr, cnt_dat, buf[]) ''<HIDE>
  Dim cnt_i, ftc As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - cnt_dat))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(cnt_dat - 1, 4)))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop
  For cnt_i = 1 To cnt_dat
    out(00001010h + (channel - 1) * 16, buf[cnt_i])         ' data count
  Next cnt_i
  Sub_End_critical_section
EndSub

'---------------------------------------------------------------
' Function only For internal use!
' send LS-Bus broadcast no acknowlege (one Byte)
' module  : module address
' channel : number of the LS interface on the LS module
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
' buf[]   : array
'---------------------------------------------------------------
Sub I_Send_BC8(channel, reg_adr,value) ''<HIDE>
  Dim ftc As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < (28 - 1))
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(1 - 1, 4)))  ' Header 1
  nop
  out(00001010h + (channel - 1) * 16, (80h Or reg_adr))     ' Header 2
  nop

  out(00001010h + (channel - 1) * 16, value)         ' data count

  Sub_End_critical_section
EndSub



'---------------------------------------------------------------
' Request to read data
' channel : number of the LS interface on the LS module
' dev_adr : device nr lsbus (1..15)
' reg_adr : adress of register (JX/FPGA)
' cnt_dat : count of data  (1..16)
'---------------------------------------------------------------
Sub LS_Read_Req(channel, ls_module, reg_adr, cnt_dat) ''<HIDE>
  Dim ftc As Long

  Do
    ftc = IN(00001015h + (channel - 1) * 16)                ' tx FIFO count
  Until (ftc < 28)
  Sub_begin_critical_section
  out(00001010h + (channel - 1) * 16, (Shift_Left(cnt_dat - 1, 4) Or ls_module))  'header 1
  nop
  out(00001010h + (channel - 1) * 16, reg_adr)              ' Header 2
  nop
  Sub_End_critical_section
EndSub

'---------------------------------------------------------------
' Read response (only DIG_IO request)
' channel : number of the LS interface on the LS module
' return  : state of digital outputs and inputs
'---------------------------------------------------------------
Function LS_Read_Resp_DIO(channel) As Long ''<HIDE>
  Dim ld[4] As Long At DM_Local
  Dim cnt_i As Long

  For cnt_i = 1 To 4                                        ' rx data read
    ld[cnt_i] = IN(00001010h + (channel - 1) * 16)          ' rx data read
  Next cnt_i
  LS_Read_Resp_DIO = ld[1] Or (Shift_Left(ld[2], 8)) Or (Shift_Left(ld[3], 16)) Or (Shift_Left(ld[4], 24))
EndFunction

'---------------------------------------------------------------
' Read response
' channel : number of the LS interface on the LS module
' cnt_dat : count of data  (1..16)
' buf[]   : array (min len is cnt_dat)
' return  : error code (0->ok)
'---------------------------------------------------------------
Function LS_Read_Resp(channel, cnt_dat, buf[]) As Long ''<HIDE>
  Dim cnt_i As Long
  For cnt_i = 1 To cnt_dat                                  ' rx data read
    buf[cnt_i] = IN(00001010h + (channel - 1) * 16)         ' rx data read
  Next cnt_i
  LS_Read_Resp = IN(00001010h + (channel - 1) * 16)         ' acknowledge
EndFunction

'******************************************************************************
'''Resets the watchdog timer, sets digital outputs and
'''+reads digital inputs.
'''+Return: state of digital outputs and inputs
'''channel: number of the LS interface on the LS module(1..2)
'''+ ls_module: device nr lsbus (FIX 1)
'''value: each bit corresponds to a channel
'''****************************************************************************
Function LS_Dig_IO(channel, value) As Long

  Dim cnt_i As Long
  dim invalue as long
  invalue=0
  For cnt_i = 0 To 24 step 8           
    invalue=invalue OR Shift_left( IN(00001010h + (channel - 1) * 16),cnt_i)          ' rx data read
  Next cnt_i
  LS_Dig_IO = invalue
                                               
  I_Send_BC8(channel, 44h, 1) ' WD trigger

  I_Send_NA32(channel, 1, 7Ch,value)                         ' channel 1..32
  LS_Read_Req(channel, 1, 7Ch, 4)                           ' request new in data
EndFunction

'******************************************************************************
'''Enables, disables and configures the watchdog.
'''+Return: error code (0->ok)
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr lsbus (1..15)
'''enable:  
'''+           0->disable
'''+           1->enable
'''time:  trigger time [ms]
'''****************************************************************************
Function LS_Watchdog_Init(channel, ls_module, enable, time) As Long
  Dim ld[2] As Long
  Dim ret, t As Long

  ld[1] = 0ADh   'Watchdog disable
  ld[2] = 0
  ret = I_Send(channel, ls_module, 45h, 2, ld)
  'Watchdog time
  t = time * 40000                                          't = time / 25e-6

  ret = ret Or (I_Send32(channel, ls_module, 48h,t))
  ' Watchdog enable
  ld[1] = 0ADh
  ld[2] = enable And 1
  LS_Watchdog_Init = ret Or (I_Send(channel, ls_module, 45h, 2, ld))
EndFunction

'******************************************************************************
'''Resets the watchdog timer.
'''+If the watchdog is activated, the user program has to call
'''+this Function at least every ....
'''channel: number of the LS interface on the LS module (1..2)
'''****************************************************************************
Sub LS_Watchdog_Reset(channel)
  I_Send_BC8(channel, 44h, 1)
EndSub

'******************************************************************************
'''Initialises a digital I/O module (channel 0 - 31).
'''+Return: error code (0->ok)
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''value: each bit corresponds to 8 channel
'''+           0 -> input
'''+           1 -> output
'''+           bit 0: channels 1..8
'''+               1: channels 9..16
'''+               2: channels 17..24
'''+               3: channels 25..32
'''****************************************************************************
Function LS_DigProg(channel, ls_module, value) As Long
  Dim ld[2] As Long At DM_Local
  Dim ret As Long

  ld[1] = (Shift_Left(value, 3) And 8h) OR (Shift_Left(value, 7) And 80h) OR 11H
  ld[2] = (Shift_Left(value, 2) And 8h) OR (Shift_Left(value, 6) And 80h) OR 11H
  ret = I_Send(channel, ls_module, 16h, 2, ld)
  ld[1] = (Shift_Left(value, 1) And 8h) Or (Shift_Left(value, 5) And 80h) OR 11H
  ld[2] = (value And 8h) OR (Shift_Left(value, 4) And 80h) OR 11H
  LS_DigProg = ret Or (I_Send(channel, ls_module, 20h + 16h, 2, ld))
EndFunction

'******************************************************************************
'''Sets the digital outputs (channel 0 - 31).
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''value: each bit corresponds to a channel
'''****************************************************************************
Sub LS_Digout_Long(channel, ls_module, value)
  I_Send_NA32(channel, ls_module, 7Ch, value)                 ' channel 1..32
EndSub

'******************************************************************************
'''Sets the digital outputs (channel 0 - 31).
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''value: each bit corresponds to a channel
'''BUS_stat: Error Code 0=OK 
'''****************************************************************************
Sub LS_Digout_Long_BS(channel, ls_module, value,BUS_Stat)
  BUS_Stat=I_Send32(channel, ls_module, 7Ch, value)                 ' channel 1..32
EndSub

'******************************************************************************
'''Reads the digital inputs (channel 0 - 31).
'''+Return: state of digital outputs and inputs
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''****************************************************************************
Function LS_Digin_Long(channel, ls_module) As Long
  Dim ret As Long
  ret = I_Receive32(channel, ls_module, 7Ch,LS_Digin_Long )
EndFunction

'******************************************************************************
'''Reads the digital inputs (channel 0 - 31).
'''+Return: state of digital outputs and inputs
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''BUS_stat: Error Code 0=OK 
'''****************************************************************************
Function LS_Digin_Long_BS(channel, ls_module,BUS_Stat) As Long
  BUS_Stat = I_Receive32(channel, ls_module, 7Ch,LS_Digin_Long_BS)
EndFunction

'******************************************************************************
'''Reads the digital output status (overcurrent) (channel 0 - 31)
'''+Return: each bit corresponds a channel 0 = OK 1 = Error
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr lsbus
'''****************************************************************************
Function LS_Get_Output_Status(channel, ls_module) As Long
  Dim t As Long
  t = I_Receive32(channel, ls_module, 50h, LS_Get_Output_Status)
EndFunction

'******************************************************************************
'''Initialises a LS-Module.
'''+Return: error code (0->ok)
'''channel: number of the LS interface on the LS module(1..2)
'''ls_module: device nr LS-bus (1..15)
'''****************************************************************************
Function LS_DIO_Init(channel, ls_module) As Long
  Dim ld[4] As Long At DM_Local
  Dim ret As Long
  Dim addr As Long
  
  out(00001013h + (channel - 1) * 16, 1)                    ' ' autoreset rx ON
  ld[1] = 3
  ret = I_Send(channel, ls_module, 40h, 1, ld)              ' Reset JX 1, 2
  ld[1] = 1
  ret = ret Or (I_Send(channel, ls_module, 5Ch, 1, ld))     ' timeout reset
  ld[1] = 4
  ret = ret Or (I_Send(channel, ls_module, 4Ch, 1, ld))     ' Set default direction time
  ld[1] = 0
  ld[2] = 0
  ld[3] = 0
  ld[4] = 0
  LS_DIO_Init = ret Or (I_Send(channel, ls_module, 50h, 4, ld))  ' overcurrent reset
  ld[1] = 30h
  LS_DIO_Init = ret Or (I_Send(channel, ls_module, 43h, 1, ld))  ' overtemperature reset
EndFunction

'******************************************************************************
'''Reset the LS - Controller
'''channel: number of the LS interface on the LS module(1..2)
'''****************************************************************************
SUB LS_Reset(channel) 
  out(00001011h + (channel - 1) * 16, 11B)                    'clear rx und tx
  sleep(3)
ENDSUB


'<ENDREGION>


